/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 2000 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: radu@eecs.umich.edu
*******************************************************************/

package dacia;

import java.io.*;
import java.util.*;
import java.net.*;


/**
 * The engine is the heart of DACIA. It decouples the application and
 * component-specific code and functionality from the general administrative
 * tasks such as maintaining the list of PROCs and their connections,
 * migrating PROCs, establishing and maintaining connections between hosts
 * and communicating between hosts.
 * @see Proc
 * @see Connection
 * @see Message
 * @author Radu Litiu
 */
public class Engine {
  /**
   * NetworkMessage types
   */
  static final int DATA             = 0;
  static final int PROCLIST         = 1;
  static final int DISCONNECT       = 2;
  static final int CONNECT_PROCS    = 3;
  static final int DISCONNECT_PROCS = 4;
  static final int MOVE             = 5;
  static final int UPDATE           = 6;
  static final int LOAD				= 7;  
  static final int CLASS			= 8;

  /** local host identification */
  static String hostname = null;
  /** fully qualified host name */
  static String fullHostname = null;
  /** used to generate an ID for a newly created PROC */
  static int nextProcID = 0;

  /** port to listen to connection requests from other engines */
  static int port = 5201;
  private static ServerSocket serverSocket = null;
  /** the thread that accepts connections from other engines */
  private static EngineThread listener = null;
  /** timeout interval for re-establishing broken connections */
  private static final int TIMEOUT = 7000;

  /** 
   * The port where the broker runs<p>
   * This is unused in this version.
   */
  static int brokerPort = 5203;
  private static Socket brokerSocket = null;
  private static ObjectInputStream ois = null;
  private static ObjectOutputStream oos = null;

  /** The local PROCs */
  static Hashtable localProcs = new Hashtable(10);
  /** The remote PROCs */
  static Hashtable remoteProcs = new Hashtable(10);
  /** total number of PROCs */
  static int procNo = 0;
  /** The other Engines - unused in this version */
  static Hashtable engines = new Hashtable(10);
  /** The Connections to other Engines */
  static Vector connections = new Vector(10, 10);
  /** The thread that monitors the system and performs adaptation */
  static Monitor monitor = null;
  /** flag used to prevent updating the info about remote PROCs when a connect to a remote host is caused by a connectProc/disconnectProc */
  static boolean updateProcs = true;

  /** the AWT panel that displays application and PROC info */
  static Graph graph;
  static boolean updateFlag = false;
  /* used to time a PROC move
  static long tstart = 0;
  static int cnt = 0;
  */    
  private static DClassLoader loader;  
  private static Thread mainThread = null;  
  static byte[] classBuffer = null;

  public static void init(String configFileName) {
    readConfigFile(configFileName);
    init();
  }

  public static void init() {
    InetAddress localAddress = null;
    try{
      localAddress = InetAddress.getLocalHost();
    }
    catch(UnknownHostException e) {
      System.out.println("Engine: exception in reading local address: "
                         + e.getMessage());
      e.printStackTrace();
    }
    fullHostname = localAddress.getHostName().toLowerCase();
    int i = fullHostname.indexOf(".");
    if(i < 0) {
      hostname = fullHostname;
    }
    else {
      hostname = fullHostname.substring(0, i);
    }

    nextProcID = (localAddress.hashCode()*1000)%1000000;
    System.out.println("Engine configuration :\n\thostname = " + fullHostname +
           "\n\tport number = " + port +
           //       "\n\tbroker port number = " + brokerPort +
           "\n\tNextProcID = " + nextProcID);

    // connects to the local broker
    /*try {
      brokerSocket = new Socket("localhost", brokerPort);
      oos = new ObjectOutputStream(brokerSocket.getOutputStream());
      oos.flush();
      ois = new ObjectInputStream(brokerSocket.getInputStream());
    }
    catch (Exception e) {
      System.out.println("Engine: exception while connecting to the broker"
                         + e.getMessage());
      e.printStackTrace();
      }*/

    // listens to connections from other engines
    listener = new EngineThread(null);
    listener.start();
    /** thread used to notify a blocked thread that a message has been
     * received. Currently this is used for class loading
     */
    mainThread = Thread.currentThread();
  }

  private static void readConfigFile(String configFileName) {
    String configInfo = null;
    try {
      File configFile = new File(configFileName);
      int fileLength = (int)configFile.length();
      FileInputStream fi = new FileInputStream(configFileName);
      byte[] buf = new byte[fileLength];
      fi.read(buf, 0, fileLength);
      configInfo = new String(buf, 0, fileLength);
      fi.close();
    }
    catch(Exception e) {
      System.out.println("Engine: exception in reading file " +
                         configFileName + ": " + e.getMessage());
      e.printStackTrace();
      System.exit(1);
    }
    // if a key is not found, it is set to a default value
    StringTokenizer st = new StringTokenizer(configInfo, " #\t\n\r=", true);
    String key = null;
    while (st.hasMoreTokens()) {
      String token = st.nextToken();
      if(token.equals("#")) { // look for the end of line
        while(st.hasMoreTokens() && !st.nextToken().equals("\n")) {}
        key = null;
      }
      else if (token.equals(" ") || token.equals("=") || token.equals("\t")) {
        // do nothing
      }
      else if (token.equals("\n") || token.equals("\r")) { // reset
        key = null;
      }
      else if (token.equals("PORT_NUMBER") ||
               token.equals("BROKER_PORT_NUMBER")) {
        key = token;
      }
      else if (key != null) {// token is a key value
        if (key.equals("PORT_NUMBER")) {
          port = Integer.parseInt(token, 10);
        }
        else if (key.equals("BROKER_PORT_NUMBER")) {
          brokerPort = Integer.parseInt(token, 10);
        }
      }
      else { // unknown token -> do nothing
        System.out.println("Engine: unknown token: " + token);
      }
    }
  }

  /**
   * Accepts incoming connections from other engines.
   */
  static void listenToSocket() {
    Socket socket = null;
    Connection connection = null;
    try {
      serverSocket = new ServerSocket(port);
    }
    catch (Exception e) {
      System.out.println("Exception while creating a ServerSocket" +
                         e.getMessage());
      e.printStackTrace();
    }
    while (true) {
      try {
        socket = serverSocket.accept();
        socket.setTcpNoDelay(true);
        // socket.setSoLinger(true, 0);
        if(socket != null) {
          connection = new Connection(socket);
          addConnection(connection);
          // sends the list (Vector) of local PROCs
          NetworkMessage netmsg = new NetworkMessage(PROCLIST);
          netmsg.data = (Object)getProcs(true);
          netmsg.sendMessage(connection);
          // start a listener for this socket
          (new EngineThread(connection)).start();
        }
      }
      catch(Exception e) {
        System.out.println("Engine: exception in accept():" + e.getMessage());
        e.printStackTrace();
        return;
      }
      System.out.println("Connection accepted: " + socket.toString());
    }
  }

  /**
   * Handles messages received from another engine.
   * @param connection  the connection to listen to
   */
  static void listenToSocket(Connection connection) {   
    NetworkMessage netmsg = null;
    Message msg = null;
    while(true) {
      try {
        netmsg = new NetworkMessage(connection);
        //System.out.println("Received message of type " + netmsg.type);
      }
      catch(EOFException e) {
        // the other host closed the connection
        System.out.println("Connection closed: "+connection.socket.toString());
        // e.printStackTrace();
        connectionClosed(connection);
        break;
      }
      catch(SocketException e) {
          //       if(e.getMessage().equals("Connection reset by peer")) {
	      if(e.getMessage().startsWith("Connection reset by peer")) {
          // the other host closed the connection
          System.out.println("Connection closed: "+connection.socket.toString());
          // e.printStackTrace();
          connectionClosed(connection);
          break;
        }
        else {  
          // this host closed the connection
          System.out.println(e.getMessage() +": " +connection.socket.toString());
          e.printStackTrace();
          break;
        }
      }
      catch(Exception e) {
        System.out.println("Socket: Exception while reading object:\n" +
                           e.getMessage());
        e.printStackTrace();
        // break;
      }
      if(netmsg.type == DATA) {
        output(netmsg);
      }
      else if(netmsg.type == PROCLIST) {
        if(updateProcs) {
          // send the message to the other engines
          // sendMessageToAll(netmsg, connection);
          RemoteProc proc;
          Vector procList = (Vector)netmsg.data;
          boolean isLocal = true;
          int size = procList.size();
          // remove the remote PROCs hosted by the sender of this message
          removeRemoteProcs(connection.hostname);
          for(int i=0; i<size; i++) {
            proc = (RemoteProc)procList.elementAt(i);
            if(proc == null) {
              // separator between PROCs local and remote to the sender
              isLocal = false;
              continue;
            }
            if(isLocal) {
              // PROCs local to the sender
              // System.out.println("Received local PROC: " + proc.id);
              proc.connection = connection;
              remoteProcs.put(new Integer(proc.id), proc);
              procNo++;
            }
            else {
              // PROCs remote to the sender
              // System.out.println("Received remote PROC: " + proc.id);
                //if(localProcs.get(new Integer(proc.id)) == null) {
              if(!(proc.hostname.equals(hostname))) {
                // it is not one of the local PROCs
                if(netmsg.i1 == 1) 
                  updateRemoteProc(proc, true);
                else
                  updateRemoteProc(proc, false);
              }
            }
          }
          //updateGraph();
          updateFlag = true;
        }
        else  // ignore this message, reset flag
          updateProcs = true;
      }
      else if(netmsg.type == CONNECT_PROCS) {
        //System.out.println("connectProcs received: " + netmsg.i1 + " " +  netmsg.i2 + " " + netmsg.i3 + " " + netmsg.i4 + " " + netmsg.i5);
        if(netmsg.i5 == 0)
          connectProcs(netmsg.i1, netmsg.i2, netmsg.i3, netmsg.i4, false);
        else
          connectProcs(netmsg.i1, netmsg.i2, netmsg.i3, netmsg.i4, true);
        //updateGraph();
      }
      else if(netmsg.type == DISCONNECT_PROCS) {
        // System.out.println("disconnectProcs received: " + netmsg.i1 + " " +  netmsg.i2 + " " + netmsg.i5);
        if(netmsg.i5 == 0)
          disconnectProcs(netmsg.i1, netmsg.i2, false);
        else
          disconnectProcs(netmsg.i1, netmsg.i2, true);
        //updateGraph();
      }
      else if(netmsg.type == MOVE) {
        if(netmsg.i1 == 0) // move from the remote host to this host
          receiveProc((Proc)netmsg.data);
          /* used to time a PROC move
          moveProc((Proc)netmsg.data, connection);
          */
        else if(netmsg.i1 == 1)
          // the PROC has to move from this host to another host
          moveProc(new Integer(netmsg.i2), (String)netmsg.data);
          /* this is not complete as it is. Should notify the remote engine
             issuing the move() command if it fails
          */
        else {
          // this is a PROC move notification
          // update the location of the remote PROC
          RemoteProc rp = (RemoteProc)remoteProcs.get(new Integer(netmsg.i1));
          if(rp !=null)
            rp.hostname = (String)netmsg.data;
        }
 		//updateGraph();
     }
      else if(netmsg.type == UPDATE) {
        netmsg.type = PROCLIST;
        if(netmsg.i1 == 1)  // all PROCs
          netmsg.data = (Object)getProcs(true);
        else  // local PROCs only
          netmsg.data = (Object)getProcs(false);
        netmsg.sendMessage(connection);
      }
      else if(netmsg.type == DISCONNECT) {
        disconnectHost();
      }
      else if(netmsg.type == LOAD) {
        if (netmsg.data == null)
          System.out.println("Receive a null request" );
        String filename = (String)netmsg.data;
        System.out.println("Receive a request:" + netmsg.data);
        filename = filename.replace ('.', File.separatorChar) + ".class";
		    
        // Create a file object relative to directory provided
        File f = new File (filename);
        int size = (int)f.length();
        System.out.println(filename + " " + size);

        // Reserve space to read
        byte buf[] = new byte[size];
        
        if (size > 0) {
           try {
             // Get stream to read from
             FileInputStream fis = new FileInputStream(f);
             fis.read(buf);
             fis.close();
           }
           catch (FileNotFoundException e) {
             System.out.println(e.getClass().getName());
             continue;
           }
           catch (IOException e) {
             System.out.println(e.getClass().getName());
             continue;
           }
        } 
        netmsg.type = CLASS;
        //netmsg.dataBuffer = buf;		
        netmsg.data = buf;		
        netmsg.sendMessage(connection);		
      }
      else if (netmsg.type == CLASS) {
        //System.out.println("Receive the class from remote engine");
        if (netmsg.data == null)
          System.out.println("Empty class returned from remote engine");  
        else {
          classBuffer = (byte[])netmsg.data;
          synchronized(mainThread) {
            //System.out.println("Notifying thread");
            mainThread.notify();
          }		
        }
      }	
      else {
        System.out.println("Unknown message type: " + netmsg.type);
      }
    }
  }

  /**
   * Sends a message to a PROC
   * @param netmsg  the message
   */
  private static void output(NetworkMessage netmsg) {
    Integer procID = new Integer(netmsg.i1);
    RemoteProc rp = null;
    Proc proc = (Proc)localProcs.get(procID);
    if(proc == null) {
      if((rp = (RemoteProc)remoteProcs.get(procID)) != null) {
        netmsg.sendMessage(rp.connection);
      }
      else {
        System.out.println("Can't send message: PROC " + netmsg.i1 +
                           " not found.");
      }
    }
    else if(proc.moving) {
      synchronized(proc) {
        try {
          proc.wait();
        }
        catch (Exception e) {
          System.out.println("output(): exception raised while wait(): "
                             + e.getMessage());
          e.printStackTrace();
        }
      }
      if((rp = (RemoteProc)remoteProcs.get(procID)) != null) {
        netmsg.sendMessage(rp.connection);
      }
      else {
        System.out.println("Can't send msg: PROC " + netmsg.i1 +" not found.");
      }
    }
    else {
      proc.input(netmsg.i2, (Message)netmsg.data, 0);
    }
  }


  /**
   * Connects to the Engine running on the specified hostname:port
   * @param hostname  the host to connect to
   * @param port  the number of the port the remote engine is listening to
   * @param update  if true, exchange PROC info with the remote host
   * @return the connection opened
   */
  public static Connection connect(String hostname, int port, boolean update) {
    Socket socket = null;
    Connection connection = null;
    try {
      updateProcs = update;
      /* Use InetAddress.getByName(hostname).getHostAddress() instead of
       * hostname below to get the fully qualified address.
       * Make the appropriate change in Connection.java and everywhere else
       * a hostname is used.
       */
      socket = new Socket(hostname, port);
      socket.setTcpNoDelay(true);
      // socket.setSoLinger(true, 0);
      if(socket != null) {
        connection = new Connection(socket);
        addConnection(connection);
        if(update) { // sends the list (Vector) of local PROCs
          NetworkMessage netmsg = new NetworkMessage(PROCLIST);
          netmsg.data = (Object)getProcs(true);
          netmsg.sendMessage(connection);
        }
        // start a listener for this socket
        (new EngineThread(connection)).start();
      }
    }
    catch (Exception e) {
      updateProcs = true;
      System.out.println("Can't connect to " + hostname + ":" + port + " : "
                         + e.getMessage());
      //e.printStackTrace();
      return null;
    }
    System.out.println("Connection established: " + socket.toString());
    return connection;
  }

  /**
   * Sends a message on a specified connection.
   * @param msg  the message
   * @param connection  the connection where the message should be sent
   * @param procID  the ID of the destination PROC
   * @param port  the destination port
   */
  static void sendMessage(Message msg, Connection connection, int procID,
                          int port) {
    NetworkMessage netmsg = new NetworkMessage(DATA);
    netmsg.i1 = procID;
    netmsg.i2 = port;
    netmsg.data = (Object)msg;
    netmsg.sendMessage(connection);
  }

  /**
   * Sends a message to all the connected engines.
   * @param netmsg  the message
   */
  static void sendMessageToAll(NetworkMessage netmsg) {
    Connection connection = null;
    for(int i=0; i<connections.size(); i++) {
      connection = (Connection)connections.elementAt(i);
      netmsg.sendMessage(connection);
    }
  }

  /**
   * Sends a message to all the connected engines, but the original sender
   * of the message.<p>
   * This method may not be needed in the current version.
   * @param netmsg  the message
   * @param origin  the connection where this message originated, to be excluded from the destination list
   */
  static void sendMessageToAll(NetworkMessage netmsg, Connection origin) {
    Connection connection = null;
    for(int i=0; i<connections.size(); i++) {
      connection = (Connection)connections.elementAt(i);
      if(connection != origin) {
        netmsg.sendMessage(connection);
      }
    }
  }

  /**
   * Adds a connection to another engine.
   * @param connection  the connection
   */
  private static void addConnection(Connection connection) {
    connections.addElement(connection);
    updateFlag = true;
    //updateGraph();
  }

  /**
   * Instructs a host to close all its connections to other hosts.<p>
   * This method is used by some monitors.<p>
   * WARNING: use it with discretion. Don't abuse it!
   * @param connection  the connection to the host to disconnect
   */
  public static void disconnectHost(Connection connection) {
    NetworkMessage netmsg = new NetworkMessage(DISCONNECT);
    netmsg.sendMessage(connection);
    System.out.println("disconnectHost(connection) called");
  }

  /**
   * Closes all connections between this engine and other engines.
   * Remove all the remote PROCs.
   * This method is used by some monitors.
   * WARNING: use it with discretion. Don't abuse it!
   */
  private static void disconnectHost() {
    Proc proc = null;
    Enumeration procEnum = null;
    Connection connection;
    int i;
    System.out.println("disconnectHost() called");
    try{
      // prevent other engines from reconnecting
      listener.stop();
      serverSocket.close();
      System.out.println("ServerSocket closed");

      for(i=0; i<connections.size(); i++) {
        (connection = (Connection)(connections.elementAt(i))).socket.close();
        if(graph != null)
          graph.removeHost(connection.hostname);
      }
      connections.removeAllElements();
      procEnum = localProcs.elements();
      while(procEnum.hasMoreElements()) {
        proc =(Proc)procEnum.nextElement();
        for(i=0; i<proc.portNo; i++)
          if(proc.port[i].connectionStatus == 2)
            proc.port[i].connectionStatus = 0;
      }
      procNo -= remoteProcs.size();
      remoteProcs.clear();
      updateFlag = true;
      Thread.currentThread().sleep(TIMEOUT + 1000);
    }
    catch(Exception e) {
      System.out.println("Exception in disconnectHost(): " + e.getMessage());
      e.printStackTrace();
    } 
    // enable other engines to reconnect
    listener = new EngineThread(null);
    listener.start();
  }

  /**
   * Handles a broken connection to another engine.<p>Try to reconnect first.
   * If this is not possible, remove this connection.
   * @param connection  the connection
   */
  private static void connectionClosed(Connection connection) {
    String hostname = connection.hostname;
    InetAddress adr = connection.socket.getInetAddress();
    Socket s = null;
    RemoteProc rp;
    int i=500, time=0;
    // attemp to reconnect, after a timeout interval
    while((time < TIMEOUT) && (s == null)) {
      try {
        i *= 2;
        time += i;
        Thread.currentThread().sleep(i);
        // s = new Socket(hostname, connection.socket.getPort());
        //s = new Socket(hostname, port);
        s = new Socket(adr, port);
      }
      catch(IOException e) {
        // System.out.println("IOException in connectionClosed: " + e.getMessage());
        // e.printStackTrace();
        s = null;
      }
      catch(Exception e) {
        System.out.println("Exception in connectionClosed: " + e.getMessage());
        e.printStackTrace();
      }
    }
    if(s != null) {
      // the connection with the remote host has been re-established
      // restore the socket and the I/O streams for this connection
      System.out.println("Connection to host " + hostname + " re-established");
      connection.socket = s;
      try{
        connection.dos = new DataOutputStream(s.getOutputStream());
        connection.dis = new DataInputStream(s.getInputStream());
        NetworkMessage netmsg = new NetworkMessage(PROCLIST);
        netmsg.data = (Object)getProcs(true);
        netmsg.sendMessage(connection);
        (new EngineThread(connection)).start();
      }
      catch(IOException e) {
        System.out.println("Exception in connectionClosed: " + e.getMessage());
        e.printStackTrace();
      }
    }
    else {
      // remove the connection and the PROCs located on the failed host
      // allow for outstanding structural changes to complete
      System.out.println("Can't re-establish connection to host " + hostname);
      try {
        Thread.currentThread().sleep(300);
      }
      catch(Exception e) { }
      Enumeration procEnum = remoteProcs.elements();
      while(procEnum.hasMoreElements()) {
        if((rp =(RemoteProc)procEnum.nextElement()).hostname.equals(hostname)){
          for(i=0; i<rp.portNo; i++)
            if(rp.connectedProcID[i] != 0)
              disconnectProcs(rp, i, false);
          remoteProcs.remove(new Integer(rp.id));
          procNo--;
          //System.out.println("PROC removed: " + rp.id);
        }
      }
    connections.removeElement(connection);
    if(graph != null)
      graph.removeHost(connection.hostname);
    updateFlag = true;
    }
  }

  /**
   * Adds a local PROC.
   * @param proc  the PROC
   */
  public static void addProc(Proc proc) {
    localProcs.put(new Integer(proc.id), proc);
    procNo++;
    updateFlag = true;
    //updateGraph();
  }

  /**
   * Removes a PROC.
   * @param proc  the PROC
   */
  public static boolean removeProc(Integer procID) {
    Proc proc = null;
    Port port1 = null, port2 = null;

    if ((proc = (Proc)localProcs.get(procID)) != null) {
      /* update the connections of the local PROCS to the moving PROC and
       * cut the connections of the moving PROC
       */
      for(int i=0; i<proc.portNo; i++) {
        if((port1 = proc.port[i]).connectionStatus == 1) {
          port2 = port1.connectedProc.port[port1.connectedPortNo];
          port2.connectionStatus = 0;
          port2.connectedPortNo = 0;
          port2.connectedProcID = 0;
          port2.connectedProc = null;
        }
        port1.myProc = null;
        port1.connectedProc = null;
        port1.remoteConnection = null;
      }
      removeLocalProc(proc);
      graph.removeNode(proc.id);
      updateFlag = true;
      return true;
    }
    else {
      System.out.println("removeProc: cannot remove a remote Proc");
      return false;
    }
  }

  /**
   * Removes a local PROC.
   * @param proc  the PROC
   */
  public static void removeLocalProc(Proc proc) {
    localProcs.remove(new Integer(proc.id));
    procNo--;
    updateFlag = true;
    //updateGraph();
  }

  /**
   * Removes all the remote PROCs located on the host indicated
   * @param hostname  the host where the PROCs to be removed are located
   */
  private static void removeRemoteProcs(String hostname) {
    RemoteProc rp;
    Enumeration procEnum = remoteProcs.elements();
    while(procEnum.hasMoreElements())
      if((rp =(RemoteProc)procEnum.nextElement()).hostname.equals(hostname)) {
        remoteProcs.remove(new Integer(rp.id));
        procNo--;
      }
    updateFlag = true;
    //updateGraph();
  }

  /**
   * Updates a remote PROC or adds it if it does not exist.
   * @param proc  the remote PROC
   * @param update  if true, then the PROC to be added overwrites an existing remote PROC with the same ID, if applicable
   */
  private static void updateRemoteProc(RemoteProc proc, boolean update) {
    Integer procID = new Integer(proc.id);
    if(remoteProcs.get(procID) == null) {
      // PROC not found
      // System.out.println("Add remote PROC: " + proc.id);
      remoteProcs.put(procID, proc);
      procNo++;
    }
    else if(update) {
      // replace the previously existing remote PROC with the new one
      // System.out.println("Update remote PROC: " + proc.id);
      remoteProcs.remove(procID);
      remoteProcs.put(procID, proc);
    }
    //updateFlag = true;
    //updateGraph();
  }

  /**
   * Connects two PROCs, local or remote.
   * @param procID1  the first PROC
   * @param port1  the first port
   * @param procID2  the second PROC
   * @param port2  the second port
   * @param propagate  if true, inform the other engines about this connection
   * @return true  if connection is successful
   */
  public static boolean connectProcs(int procID1, int port1, int procID2,
                                     int port2, boolean propagate) {
    Proc p1=null, p2=null;
    RemoteProc p3=null, p4=null;
    if((p1 = (Proc)localProcs.get(new Integer(procID1))) != null) {
      if((p2 = (Proc)localProcs.get(new Integer(procID2))) != null) {
        return connectProcs(p1, port1, p2, port2, propagate);
      }
      else if((p4 = (RemoteProc)remoteProcs.get(new Integer(procID2)))!= null){
        return connectProcs(p1, port1, p4, port2, propagate);
      }
      else {
        System.out.println("Connect1: Can't find PROC " + procID2);
        return false;
      }
    }
    else if((p3 = (RemoteProc)remoteProcs.get(new Integer(procID1))) != null){
      if((p2 = (Proc)localProcs.get(new Integer(procID2))) != null) {
        return connectProcs(p3, port1, p2, port2, propagate);
      }
      else if((p4 = (RemoteProc)remoteProcs.get(new Integer(procID2))) !=null){
        return connectProcs(p3, port1, p4, port2, propagate);
      }
      else {
        System.out.println("Connect2: Can't find PROC " + procID2);
        return false;
      }
    }
    else {
      System.out.println("Connect3: Can't find PROC " + procID1);
      return false;
    }
  }

  /**
   * Connects two local PROCs.
   * @param proc1 the first PROC - local
   * @param port1  the first port
   * @param proc2 the second PROC - local
   * @param port2  the second port
   * @param propagate  if true, inform the other engines about this connection
   * @return true  if connection is successful
   */
  public static boolean connectProcs(Proc proc1, int port1, Proc proc2,
                                     int port2, boolean propagate) {
    // the connecting port is not specified
    if(port1 < 0) {
      if(port2 <0) {
        System.out.println("Connect procs: " + proc1.id + " to " + proc2.id + 
                             ". At least one port should be specified");
          return false;
      }
      synchronized(proc1) {
        if((port1=proc1.getFreePort()) >= 0)
          return connectProcs(proc1, port1, proc2, port2, propagate);
        else {
          System.out.println("Can't connect proc " + proc1.id +
                             ". No port available.");
          return false;
        }
      }
    }
    if(port2 < 0) {
      synchronized(proc2) {
        if((port2=proc2.getFreePort()) >= 0)
          return connectProcs(proc1, port1, proc2, port2, propagate);
        else {
          System.out.println("Can't connect proc " + proc2.id +
                             ". No port available.");
          return false;
        }
      }
    }

    if(proc1.port[port1].connectionStatus != 0) {
      System.out.println("Can't connect proc " + proc1.id + ":" + port1 +
                         ". Already connected.");
      return false;
    }
    if(proc2.port[port2].connectionStatus != 0) {
      System.out.println("Can't connect proc " + proc2.id + ":" + port2 +
                         ". Already connected.");
      return false;
    }
    proc1.port[port1].connect(proc2, port2, true);
    proc2.port[port2].connect(proc1, port1, false);
    if(propagate) {
      NetworkMessage netmsg = new NetworkMessage(CONNECT_PROCS);
      netmsg.i1 = proc1.id;
      netmsg.i2 = port1;
      netmsg.i3 = proc2.id;
      netmsg.i4 = port2;
      sendMessageToAll(netmsg);
    }
    updateFlag = true;
    //updateGraph();
    return true;
  }

  /**
   * Connects two PROCs, one local and one remote. If there is no conection
   * to the remote host, a connection is opened.
   * @param proc1 the first PROC - local
   * @param port1  the first port
   * @param proc2 the second PROC - remote
   * @param port2  the second port
   * @param propagate  if true, inform the other engines about this connection
   * @return true  if connection is successful
   */
  public static boolean connectProcs(Proc proc1, int port1, RemoteProc proc2,
                                  int port2, boolean propagate) {
    Vector proclist = null;
    // the connecting port is not specified
    if(port1 < 0) {
      if(port2 <0) {
        System.out.println("Connect procs: " + proc1.id + " to " + proc2.id + 
                             ". At least one port should be specified");
          return false;
      }
      synchronized(proc1) {
        if((port1=proc1.getFreePort()) >= 0)
          return connectProcs(proc1, port1, proc2, port2, true);
         else {
          System.out.println("Can't connect proc " + proc1.id +
                             ". No port available.");
          return false;
        }
      }
    }

    if(port2 >= 0) {
      if(proc1.port[port1].connectionStatus != 0) {
        System.out.println("Can't connect proc " + proc1.id + ":" + port1 +
                           ". Already connected.");
        return false;
      }
      if(proc2.connectedProcID[port2] != 0) {
        System.out.println("Can't connect proc " + proc2.id + ":" + port2 +
                           ". Already connected.");
        return false;
      }
      if(propagate)
        proclist = getProcs(true);
      proc1.port[port1].connect(proc2.id, port2, proc2.connection, true);
      proc2.connectedProcID[port2] = proc1.id;
      proc2.connectedPortNo[port2] = port1;
    }

    if(propagate || (port2 < 0)) {
      NetworkMessage netmsg = new NetworkMessage(CONNECT_PROCS);
      netmsg.i1 = proc1.id;
      netmsg.i2 = port1;
      netmsg.i3 = proc2.id;
      netmsg.i4 = port2;
      if(port2 < 0)
        netmsg.i5 = 1;
      if(proc2.connection == null)
        proc2.connection = getConnection(proc2.hostname);
      if(proc2.connection == null) {
        // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine 
        if(port2 >= 0) { // don't mess up local updates w/ remote updates
          proc2.connection = connect(proc2.hostname, port, false);
          if(proc2.connection == null) {
            System.out.println("connectProcs error. Can't connect to host " +
                               proc2.hostname);
            return false;
          }
          // send local proclist to the remote host
          NetworkMessage netmsg1 = new NetworkMessage(PROCLIST);
          netmsg1.data = (Object)proclist;
          netmsg1.sendMessage(proc2.connection);
          // propagate connectProcs to the remote host
          //netmsg.sendMessage(proc2.connection);
          // CHANGE: propagate connectProcs to the remote host and all
          // other connected hosts
          sendMessageToAll(netmsg);
          // request updates from the remote host
          netmsg1 = new NetworkMessage(UPDATE);
          netmsg1.i1 = 1;
          netmsg1.sendMessage(proc2.connection);
          updateFlag = true;
          //updateGraph();
          return true;
        }
        else {
          proc2.connection = connect(proc2.hostname, port, true);
          if(proc2.connection == null) {
            System.out.println("connectProcs error. Can't connect to host " +
                               proc2.hostname);
            return false;
          }
        }
      }
      // CHANGE: propagate connectProcs to the remote host and all
      // other connected hosts
      // netmsg.sendMessage(proc2.connection);
      sendMessageToAll(netmsg);
    }
    if(port2 >= 0) {
      updateFlag = true;
      //updateGraph();
      return true;
    }
    else
      return false;
  }

  /**
   * Connects two PROCs, one local and one remote. If there is no conection
   * to the remote host, a connection is opened.
   * @param proc1 the first PROC - remote
   * @param port1  the first port
   * @param proc2 the second PROC - local
   * @param port2  the second port
   * @param propagate  if true, inform the other engines about this connection
   * @return true  if connection is successful
   */
  public static boolean connectProcs(RemoteProc proc1, int port1, Proc proc2,
                                  int port2, boolean propagate) {
    Vector proclist = null;
    // the connecting port is not specified
    if(port2 < 0) {
      if(port1 <0) {
        System.out.println("Connect procs: " + proc1.id + " to " + proc2.id + 
                             ". At least one port should be specified");
          return false;
      }
      synchronized(proc2) {
        if((port2=proc2.getFreePort()) >= 0)
          return connectProcs(proc1, port1, proc2, port2, true);
         else {
          System.out.println("Can't connect to proc " + proc2.id +
                             ". No port available.");
          return false;
        }
      }
    }

    if(port1 >= 0) {
      if(proc1.connectedProcID[port1] != 0) {
        System.out.println("Can't connect proc " + proc1.id + ":" + port1 +
                           ". Already connected.");
        return false;
      }
      if(proc2.port[port2].connectionStatus != 0) {
        System.out.println("Can't connect proc " + proc2.id + ":" + port2 +
                           ". Already connected.");
        return false;
      }
      if(propagate)
        proclist = getProcs(true);
      proc1.connectedProcID[port1] = proc2.id;
      proc1.connectedPortNo[port1] = port2;
      proc2.port[port2].connect(proc1.id, port1, proc1.connection, false);
    }

    if(propagate || (port1 < 0)) {
      NetworkMessage netmsg = new NetworkMessage(CONNECT_PROCS);
      netmsg.i1 = proc1.id;
      netmsg.i2 = port1;
      netmsg.i3 = proc2.id;
      netmsg.i4 = port2;
      if(port1 < 0)
        netmsg.i5 = 1;
      if(proc1.connection == null)
        proc1.connection = getConnection(proc1.hostname);
      if(proc1.connection == null) {
        // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine 
        if(port1 >= 0) { // don't mess up local updates w/ remote updates
          proc1.connection = connect(proc1.hostname, port, false);
          if(proc1.connection == null) {
            System.out.println("connectProcs error. Can't connect to host " +
                               proc1.hostname);
            return false;
          }
          // send local proclist to the remote host
          NetworkMessage netmsg1 = new NetworkMessage(PROCLIST);
          netmsg1.data = (Object)proclist;
          netmsg1.sendMessage(proc1.connection);
          // propagate connectProcs to the remote host
          //netmsg.sendMessage(proc1.connection);
          // CHANGE: propagate connectProcs to the remote host and all
          // other connected hosts
          sendMessageToAll(netmsg);
          // request updates from the remote host
          netmsg1 = new NetworkMessage(UPDATE);
          netmsg1.i1 = 1;
          netmsg1.sendMessage(proc1.connection);
          updateFlag = true;
          //updateGraph();
          return true;
        }
        else {
          proc1.connection = connect(proc1.hostname, port, true);
          if(proc1.connection == null) {
            System.out.println("connectProcs error. Can't connect to host " +
                               proc1.hostname);
            return false;
          }
        }
      }
      // CHANGE: propagate connectProcs to the remote host and all
      // other connected hosts
      // netmsg.sendMessage(proc1.connection);
      sendMessageToAll(netmsg);
    }
    if(port1 >= 0) {
      updateFlag = true;
      //updateGraph();
      return true;
    }
    else
      return false;
  }

  /**
   * Connects two remote PROCs. Propagate operation to the corresponding
   * hosts where the PROCs reside.
   * @param proc1 the first PROC - remote
   * @param port1  the first port
   * @param proc2 the second PROC - remote
   * @param port2  the second port
   * @return true  if connection is successful
   */
  public static boolean connectProcs(RemoteProc proc1, int port1,
                        RemoteProc proc2, int port2, boolean propagate) {
    Vector proclist = null;
    // the connecting port is not specified
    if(port1 < 0) {
      if(port2 < 0) {
        System.out.println("Connect procs: " + proc1.id + " to " + proc2.id + 
                             ". At least one port should be specified");
          return false;
      }
    }
    else
      if(port2 >= 0) {
        if(proc1.connectedProcID[port1] != 0) {
          System.out.println("Can't connect proc " + proc1.id + ":" + port1 +
                             ". Already connected.");
          return false;
        }
        if(proc2.connectedProcID[port2] != 0) {
          System.out.println("Can't connect proc " + proc2.id + ":" + port2 +
                             ". Already connected.");
          return false;
        }
        proclist = getProcs(true);
        if(!propagate) {
          proc1.connectedProcID[port1] = proc2.id;
          proc1.connectedPortNo[port1] = port2;
          proc2.connectedProcID[port2] = proc1.id;
          proc2.connectedPortNo[port2] = port1;
          updateFlag = true;
          //updateGraph();
          return true;
        }
        // otherwise the local update is done after the connection
        // notification is received from the remote host(s)
      }
    NetworkMessage netmsg = new NetworkMessage(CONNECT_PROCS);
    netmsg.i1 = proc1.id;
    netmsg.i2 = port1;
    netmsg.i3 = proc2.id;
    netmsg.i4 = port2;
    netmsg.i5 = 1;
    if(port1 >= 0) {
      if(port2 >= 0) {
        if(proc1.connection == null)
          proc1.connection = getConnection(proc1.hostname);
        if(proc2.connection == null)
          proc2.connection = getConnection(proc2.hostname);
        if(proc1.connection == null) {
          if(proc2.connection == null) {
            // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine. A local update has been made, so don't mess up local updates w/ remote updates (?)
            proc1.connection = connect(proc1.hostname, port, false);
            if(proc1.connection == null) {
              System.out.println("connectProcs error. Can't connect to host "
                                 + proc1.hostname);
              return false;
            }
            // send local proclist to the remote host
            NetworkMessage netmsg1 = new NetworkMessage(PROCLIST);
            netmsg1.data = (Object)proclist;
            netmsg1.sendMessage(proc1.connection);
            // propagate connectProcs to the remote host
            netmsg.sendMessage(proc1.connection);
            // request updates from the remote host
            netmsg1 = new NetworkMessage(UPDATE);
            netmsg1.i1 = 1;
            netmsg1.sendMessage(proc1.connection);
          }
          else  // send to proc2
            netmsg.sendMessage(proc2.connection);
        }
        else  // send to proc1
          netmsg.sendMessage(proc1.connection);
      }
      else {  // send to proc2
        if(proc2.connection == null)
          proc2.connection = getConnection(proc2.hostname);
        if(proc2.connection == null)
          // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine 
          proc2.connection = connect(proc2.hostname, port, true);
        if(proc2.connection == null) {
          System.out.println("error connectProcs. Can't connect to host "
                             + proc2.hostname);
          return false;
        }
        else
          netmsg.sendMessage(proc2.connection);
      }
    }
    else { // send to proc1
      if(proc1.connection == null)
        proc1.connection = getConnection(proc1.hostname);
      if(proc1.connection == null)
        // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine 
        proc1.connection = connect(proc1.hostname, port, true);
      if(proc1.connection == null) {
        System.out.println("error connectProcs. Can't connect to host "
                           + proc1.hostname);
        return false;
      }
      else
        netmsg.sendMessage(proc1.connection);
    }
    return true;
  }

  /**
   * Disconnects two PROCs, local or remote.
   * @param procID the first PROC
   * @param portNo  the first port
   * @param propagate  if true, inform the other engines about
   * the disconnection
   */
  public static boolean disconnectProcs(int procID, int portNo, 
                                        boolean propagate) {
    Proc proc;
    RemoteProc rp;
    if((proc = (Proc)localProcs.get(new Integer(procID))) != null) {
      if(proc.port[portNo].connectionStatus != 0) {
        return disconnectProcs(proc, portNo, propagate);
      }
      else {
        System.out.println("Disconnect1: PROC " + procID + ":" + portNo +
                           " is not connected");
        return false;
      }
    }
    else if((rp = (RemoteProc)remoteProcs.get(new Integer(procID))) != null){
      if(rp.connectedProcID[portNo] != 0) {
        return disconnectProcs(rp, portNo, propagate);
      }
      else {
        System.out.println("Disconnect1: PROC " + procID + ":" + portNo +
                           " is not connected");
        return false;
      }
    }
    else {
      System.out.println("Disconnect1: Can't find PROC " + procID);
      return false;
    }
  }

  /**
   * Disconnects two PROCs.
   * @param proc1 the first PROC - local
   * @param portNo1  the first port
   * @param propagate  if true, inform the other engines about
   * the disconnection
   */
  public static boolean disconnectProcs(Proc proc1, int portNo1,
                                     boolean propagate) {
    RemoteProc rp2;
    Port port1;
    NetworkMessage netmsg;
    if((port1 = proc1.port[portNo1]).connectionStatus == 1) {
      port1.connectedProc.port[port1.connectedPortNo].disconnect();
      port1.disconnect();
      if(propagate) {
        netmsg = new NetworkMessage(DISCONNECT_PROCS);
        netmsg.i1 = proc1.id;
        netmsg.i2 = portNo1;
        sendMessageToAll(netmsg);
      }
      updateFlag = true;
      //updateGraph();
      return true;
    }
    else if(port1.connectionStatus == 2) {
      if((rp2 =(RemoteProc)remoteProcs.get(new Integer(port1.connectedProcID)))
         != null) {
        rp2.connectedProcID[port1.connectedPortNo] = 0;
        rp2.connectedPortNo[port1.connectedPortNo] = 0;
        if(propagate) {
          // open a connection to the remote host, if needed
          if(rp2.connection == null)
            rp2.connection = getConnection(rp2.hostname);
          netmsg = new NetworkMessage(DISCONNECT_PROCS);
          netmsg.i1 = proc1.id;
          netmsg.i2 = portNo1;
          // if(rp2.connection != null) {
          //   netmsg.sendMessage(rp2.connection);
          sendMessageToAll(netmsg);
        }
      }
      port1.disconnect();
      updateFlag = true;
      //updateGraph();
      return true;
    }
    else {
      System.out.println("Disconnect2: PROC " + proc1.id +" is not connected");
      return false;
    }
  }

  /**
   * Disconnects two PROCs.
   * @param rp1 the first PROC - remote
   * @param portNo1  the first port
   * @param propagate  if true, inform the other engines about the disconnection
   */
  public static boolean disconnectProcs(RemoteProc rp1, int portNo1,
                                     boolean propagate) {
    Proc proc2 = null;
    RemoteProc rp2 = null;
    Integer int2;
    Vector proclist = null;
    if(rp1.connectedProcID[portNo1] != 0) {
      if((proc2 = (Proc)localProcs.get((int2 = new Integer(
                                   rp1.connectedProcID[portNo1])))) != null) {
        proc2.port[rp1.connectedPortNo[portNo1]].disconnect();
        rp1.connectedProcID[portNo1] = 0;
        rp1.connectedPortNo[portNo1] = 0;
        //updateGraph();
      }
      else if((rp2 = (RemoteProc)remoteProcs.get(int2)) != null) {
        //if(propagate)
        //  proclist = getProcs(true);
        if(!propagate) {
          rp2.connectedProcID[rp1.connectedPortNo[portNo1]] = 0;
          rp2.connectedPortNo[rp1.connectedPortNo[portNo1]] = 0;
          rp1.connectedProcID[portNo1] = 0;
          rp1.connectedPortNo[portNo1] = 0;
          //updateGraph();
          updateFlag = true;
          return true;
        }
      }
      // rp1.connectedProcID[portNo1] = 0;
      // rp1.connectedPortNo[portNo1] = 0;
    }
    else {
      System.out.println("Disconnect3: PROC " + rp1.id + " is not connected");
      return false;
    }
    if(propagate) {
      NetworkMessage netmsg = new NetworkMessage(DISCONNECT_PROCS);
      netmsg.i1 = rp1.id;
      netmsg.i2 = portNo1;
      if(rp1.connection == null)
        rp1.connection = getConnection(rp1.hostname);
      if(proc2 != null) { // PROC2 is local
        //netmsg.sendMessage(rp1.connection);
        sendMessageToAll(netmsg);
      }
      else { // PROC2 is remote
        // operation will bounce back, and it will be further propagated
        netmsg.i5 = 1;
        if(rp1.connection == null) {
          if(rp2.connection == null)
            rp2.connection = getConnection(rp2.hostname);
          if(rp2.connection != null) {
            // Need to establish a connection. By default, connect to the same port number as the incoming port for this engine. A local update has been made, so don't mess up local updates w/ remote updates (?)
            rp1.connection = connect(rp1.hostname, port, false);
            if(rp1.connection == null) {
              System.out.println("disconnectProcs error.Can't connect to host "
                                 + rp1.hostname);
              return false;
            }
            // send local proclist to the remote host
            proclist = getProcs(true);
            NetworkMessage netmsg1 = new NetworkMessage(PROCLIST);
            netmsg1.data = (Object)proclist;
            netmsg1.sendMessage(rp1.connection);
            // propagate disconnectProcs to the remote host
            netmsg.sendMessage(rp1.connection);
            // request updates from the remote host
            netmsg1 = new NetworkMessage(UPDATE);
            netmsg1.i1 = 1;
            netmsg1.sendMessage(rp1.connection);
          }
          else  // send to rp2
            netmsg.sendMessage(rp2.connection);
        }
        else  // send to rp1
          netmsg.sendMessage(rp1.connection);
      }
    }
    updateFlag = true;
    return true;
  }

  /**
   * Moves a PROC to another Engine.
   * @param procID  the ID of the moving PROC
   * @param host  the name of the host where the PROC is moving
   */
  public static boolean moveProc(Integer procID, String host) {
    Proc proc = null;
    RemoteProc rp = null;
    Connection connection = null;
    if((proc = (Proc)localProcs.get(procID)) != null) {
      // move a local PROC to a remote host
      if((connection = getConnection(host)) == null)
        // Need to establish a connection. By default, connect to the
        // same port number as the incoming port for this engine.
        if((connection = connect(host, port, true)) == null) {
          System.out.println("Can't move PROC " + procID +
                             ". Can't connect to host " + host);
          return false;
        }
      moveProc(proc, connection);
      return true;
    }
    else if((rp = (RemoteProc)remoteProcs.get(procID)) != null){
      // move a remote PROC
      // forward the request to the engine where the PROC is located
      if(rp.connection == null)
        if((rp.connection = getConnection(rp.hostname)) == null)
          // Need to establish a connection. By default, connect to the
          // same port number as the incoming port for this engine.
          if((rp.connection = connect(rp.hostname, port, true)) == null) {
            System.out.println("Can't move PROC " + procID +
                               ". Can't connect to host " + rp.hostname);
            return false;
          }
      NetworkMessage netmsg = new NetworkMessage(MOVE);
      // netmsg.i1 is a flag indicating that this is a forwarded move request
      netmsg.i1 = 1;
      netmsg.i2 = procID.intValue();
      netmsg.data = (Object)host;
      netmsg.sendMessage(rp.connection);
      //updateFlag = true;
      return true;
    }
    else {
      System.out.println("Can't move PROC " + procID +". PROC does not exist");
      return false;
    }
  }

  /**
   * Moves a PROC to another Engine.
   * @param proc  the moving PROC
   * @param connection  the connection to the host where the PROC is moving
   */
  public static void moveProc(Proc proc, Connection connection) {
    int i, pid = proc.id;
    Integer procID = new Integer(pid);
    Connection connection1 = null;
    RemoteProc rp = null;
    Port port1 = null, port2 = null;

    /* used to time a PROC move
    if(cnt++ == 0)
      tstart = System.currentTimeMillis();
    if(cnt == 101) {
      System.out.println("Time: " + (System.currentTimeMillis() - tstart));
      cnt = 0;
      return;
    }
    */

    proc.moving = true;
    // notify the async thread of the PROC, in order for it to stop
    synchronized(proc.msgQueue) {
      try {
        proc.msgQueue.notify();
      }
      catch (IllegalMonitorStateException e) {
        System.out.println("Exception in notify()" + e.getMessage());
        e.printStackTrace();
      }
    }
    /* allow possible senders to this PROC to learn about the movement 
     * sleep until no other messages are in traffic through this PROC
     */
    do {
      try {
        Thread.currentThread().sleep(100);
      }
      catch(InterruptedException e) {
        System.out.println("Engine interrupted while sleeping");
      }
    }
    while(proc.refcnt > 0);

    // create a RemoteProc and add it to the list of local PROCs
    rp = new RemoteProc(proc);
    rp.hostname = connection.hostname;
    rp.connection = connection;
    remoteProcs.put(procID, rp);
    /* update the connections of the local PROCS to the moving PROC and
     * cut the connections of the moving PROC
     */
    for(i=0; i<proc.portNo; i++) {
      if((port1 = proc.port[i]).connectionStatus == 1) {
        port2 = port1.connectedProc.port[port1.connectedPortNo];
        port2.remoteConnection = connection;
        port2.connectionStatus = 2;
        port2.connectedProc = null;
      }
      port1.myProc = null;
      port1.connectedProc = null;
      port1.remoteConnection = null;
    }
    proc.hostname = null;
    proc.stopWorker();
    proc.pack();
    // send the moving PROC to the destination host
    NetworkMessage netmsg = new NetworkMessage(MOVE);
    netmsg.data = (Object)proc;
    netmsg.sendMessage(connection);
    // notify other engines about the PROC movement
    // assign the PROC ID to netmsg.i1, and the new PROC location to netmsg.data
    netmsg.i1 = pid;
    netmsg.data = (Object)connection.hostname;
    for(i=0; i<connections.size(); i++) {
      connection1 = (Connection)connections.elementAt(i);
      if(connection != connection1) {
        netmsg.sendMessage(connection1);
      }
    }
    // delete the moving PROC from the list of local PROCs
    localProcs.remove(procID);
    proc.moving = false;
    // wakes up the threads trying to send messages to this PROC
    synchronized(proc) {
      try {
        proc.notifyAll();
      }
      catch (IllegalMonitorStateException e) {
        System.out.println("Exception in notify()" + e.getMessage());
        e.printStackTrace();
      }
    }
    updateFlag = true;
  }

  /**
   * Receives a PROC from another Engine.
   */
  private static void receiveProc(Proc proc) {
    int i, pid = proc.id;
    Integer procID = new Integer(pid), procID1;
    Proc proc1;
    RemoteProc rp = null, rp1 = null;
    Port port, port1;

    proc.moving = false;
    proc.hostname = getHostName();
    proc.unpack();
    /* update the connections of the local PROCS to the moving PROC
     * and the connections of the moving PROC to local and remote PROCS
     */
    for(i=0; i<proc.portNo; i++) {
      if((port = proc.port[i]).connectionStatus != 0) {
        procID1 = new Integer(port.connectedProcID);
        if((proc1 = (Proc)localProcs.get(procID1)) != null) {
          port1 = proc1.port[port.connectedPortNo];
          port.connectionStatus = 1;
          port.connectedProc = proc1;
          port1.connectionStatus = 1;
          port1.connectedProc = proc;
          port1.remoteConnection = null;
        }
        else if((rp1 = (RemoteProc)remoteProcs.get(procID1)) != null) {
          port.connectionStatus = 2;
          port.remoteConnection = rp1.connection;
        }
        else {
          System.out.println("Can't connect moving PROC. PROC " +
                             port.connectedProcID + " not found.");
        }
      }
      port.myProc = proc;
    }
    // start the worker thread
    proc.startWorker();
    // add the moving PROC to the list of local PROCs
    localProcs.put(procID, proc);
    // delete the moving PROC from the list of remote PROCs
    remoteProcs.remove(procID);
    updateFlag = true;
    //updateGraph();
    //System.out.println("PROC received: " + pid);
  }    

  protected static int getNextID() {
    return nextProcID++;
  }

  public static String getHostName() {
    return hostname;
  }

  private static void printConnections() {
    System.out.println("Connections :");
    for(int i=0; i<connections.size(); i++) {
      System.out.println("    " +
                 ((Connection)connections.elementAt(i)).socket.toString());
    }
  }

  private static void printProcs() {
    Enumeration procEnum;
    System.out.println("Local PROCS :");
    procEnum = localProcs.elements();
    while (procEnum.hasMoreElements()) {
      ((Proc)procEnum.nextElement()).printInfo();
    } 
    System.out.println("Remote PROCS :");
    procEnum = remoteProcs.elements();
    while (procEnum.hasMoreElements()) {
      ((RemoteProc)procEnum.nextElement()).printInfo();
    } 
  }

  public static String connectionsInfo() {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    DataOutputStream dos = new DataOutputStream(baos);
    try {
      dos.writeBytes("Connections :\n");
      for(int i=0; i<connections.size(); i++)
        dos.writeBytes("    " +
          ((Connection)connections.elementAt(i)).socket.toString() + "\n");
      dos.flush();
      baos.flush();
    }
    catch(IOException e) {
      System.out.println("Exception in connectionsInfo()" + e.getMessage());
      e.printStackTrace();
    }
    return baos.toString();
  }

  public static String procsInfo() {
    Enumeration procEnum;
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    DataOutputStream dos = new DataOutputStream(baos);
    try {
      dos.writeBytes("Local PROCs :\n");
      procEnum = localProcs.elements();
      while(procEnum.hasMoreElements())
        ((Proc)procEnum.nextElement()).printInfo(dos);
      dos.writeBytes("Remote PROCS :\n");
      procEnum = remoteProcs.elements();
      while (procEnum.hasMoreElements())
        ((RemoteProc)procEnum.nextElement()).printInfo(dos);
      dos.flush();
    } 
    catch(IOException e) {
      System.out.println("Exception in procsInfo()" + e.getMessage());
      e.printStackTrace();
    }
    return baos.toString();
  }

  /**
   * Sets a Monitor for this application.<p>
   * The monitor needs to be subsequently started.
   */
  public static void setMonitor(Monitor m) {
    monitor = m;
  }

  /**
   * Starts the Monitor thread.
   */
  public static void startMonitor() {
    if(monitor != null) {
      monitor.start();
    }
    else {
      System.out.println("Can't start monitor. Monitor does not exist.");
    }
  }

  /**
   * Stops the Monitor thread.
   */
  public static void stopMonitor() {
    if(monitor != null) {
      monitor.stop();
    }
    else {
      System.out.println("Can't stop monitor. Monitor does not exist.");
    }
  }

  /**
   * Updates the information about the PROCs known by other engines
   * @param hostname  the name of the remote engine to get the information from, all for all engines
   * @param allProcs  1 - both PROCs local and remote to the remote host; 0 - local PROCs only
   */
  public static void update(String hostname, int allProcs) {
    Connection connection;
    NetworkMessage netmsg = new NetworkMessage(UPDATE);
    netmsg.i1 = allProcs;
    if(hostname.equals("all"))
      sendMessageToAll(netmsg);
    else {
      if((connection = getConnection(hostname)) != null)
        netmsg.sendMessage(connection);
      else if(connect(hostname, port, true) == null)
        System.out.println("Update error: can't connect to host " + hostname);
    }
  }

  /**
   * Executes the command-line shell.<p>
   * This shell provides an interface that a user or a system administrator
   * can use to manually reconfigure the appllication.
   */
  public static void runShell() {
    BufferedReader reader = new BufferedReader(
                            new InputStreamReader(System.in));
    String line = null;
    StringTokenizer st;
    String command = null, s1=null, s2=null, s3=null, s4=null;
    while(true) {
      System.out.print("Engine > ");
      try {
        line = reader.readLine();
        st = new StringTokenizer(line);
        command = st.nextToken();
        if(command.equals("help") || command.equals("h")) {
          System.out.println("\tconnect [hostname] [IPportnumber] - connects to another engine");
          System.out.println("\tconnectProcs [sourceProc] [sourcePortNo] [destProc] [destPortNo] - connects two PROCs");
          System.out.println("\tdisconnectProcs [sourceProc] [sourcePortNo] - disconnects two PROCs");
          System.out.println("\tdisplayGraph - shows a graphical interface for application management");
          System.out.println("\texit/quit - stop execution and exit");
          System.out.println("\thelp - print this help menu");
          System.out.println("\tload [classname] [hostname] - load a class. If a hostname is given, load the class from the specified host; otherwise, try to load from all known hosts");
          System.out.println("\tmove [procID] [hostname] - move a PROC to the host indicated");
          System.out.println("\tprint - print info about all the local and remote PROCs");
          System.out.println("\tremove [procID] - remove a PROC");
          System.out.println("\tstart [procID] - calls the start() method of the PROC indicated");
          System.out.println("\tstartMonitor - start the monitoring service that performs adaptation");
          //System.out.println("\tstopMonitor - start the monitoring service that performs adaptation");
          System.out.println("\tupdate all/[hostname] <all> - updates the information about PROCs known by other engines\n\t  - all(first arg) - all hosts; hostname - only the specified host\n\t  - all(second arg) - both PROCs local and remote to the remote host; nothing - local PROCs only");
        }
        else if(command.equals("connect")) {
          s1 = st.nextToken();
          s2 = st.nextToken();
          connect(s1, (new Integer(s2)).intValue(), true);
          //updateGraph();
        }
        else if(command.equals("disconnect")) {
          disconnectHost();
        }
        else if(command.equals("connectProcs")) {
          s1 = st.nextToken();
          s2 = st.nextToken();
          s3 = st.nextToken();
          s4 = st.nextToken();
          connectProcs((new Integer(s1)).intValue(),
                       (new Integer(s2)).intValue(),
                       (new Integer(s3)).intValue(),
                       (new Integer(s4)).intValue(), true);
          //updateGraph();
        }
        else if(command.equals("disconnectProcs")) {
          s1 = st.nextToken();
          s2 = st.nextToken();
          disconnectProcs((new Integer(s1)).intValue(),
                       (new Integer(s2)).intValue(), true);
          //updateGraph();
        }
        else if (command.equals("displayGraph")) {
          displayGraph();
        }
        else if(command.equals("exit") || command.equals("quit")) {
          /* it turns out that closing the sockets cause problems on some
             platforms. For instance, on Solaris 2.5, the close() call hangs.
             Simply exiting without closing the sockets works better
          for(int i=0; i<connections.size(); i++) {
            //System.out.println("Closing connection " +
            //((Connection)connections.elementAt(i)).socket.toString());
            try {
              ((Connection)connections.elementAt(i)).socket.close();
            }
            catch(IOException e) {
              System.out.println("Exception in socket close" + e.getMessage());
              e.printStackTrace();
            }
          }
           */
          System.out.println("Exiting...");
          System.exit(0);
        }
        else if (command.equals("load")) {			
          Object obj = null;
          Proc proc = null;
          s1 = st.nextToken();			
          if (st.hasMoreTokens()) 
            obj = loadClass(s1, st.nextToken());
          else 
            obj = loadClass(s1, null);
          if ((proc = (Proc)obj) != null)
            addProc(proc);		
        }
        else if(command.equals("move")) 
          moveProc(new Integer(st.nextToken()), st.nextToken());
        else if(command.equals("print")) {
          printConnections();
          printProcs();
        }
        else if(command.equals("remove")) 
          removeProc(new Integer(st.nextToken()));
        else if(command.equals("start")) {
          Proc p = null;
          if((p = (Proc)localProcs.get(new Integer(st.nextToken()))) != null) {
            p.start();
          }
          else {
            System.out.println("Can't start PROC: PROC does not exist");
          }
        }
        else if(command.equals("startMonitor")) {
          startMonitor();
        }
        else if(command.equals("stopMonitor")) {
          stopMonitor();
        }
        else if(command.equals("update")) {
          s1 = st.nextToken();
          if(st.hasMoreTokens())
            s2 = st.nextToken();
          else
            s2 = null;
          if((s2 != null) && (s2.equals("all")))
            update(s1, 1);
          else
            update(s1, 0);
          //updateGraph();
        }
        else {
          System.out.println("Engine: invalid command");
        }
      }
      catch(IOException e) {
        System.out.println("Engine: exception in reading input: " +
                           e.getMessage());
        e.printStackTrace();
      }
      catch(Exception e) {
        System.out.println("Engine: invalid input: " + e.getMessage());
      e.printStackTrace();
      }
    }
  }

  /* Various get...() methods
   */
  public static Hashtable getLocalProcs() {
    return localProcs;
  }

  public static Hashtable getRemoteProcs() {
    return remoteProcs;
  }

  public static Vector getConnections() {
    return connections;
  }

  /**
   * Returns a Vector containing a list of all the PROCs (local and remote)
   * known by this Engine. A null element is inserted between the local and
   * remote PROCs.
   * @param allProcs  specifies whether all the PROCs should be returned (if true) or only the local PROCs (if false)
   */
  private static Vector getProcs(boolean allProcs) {
    Vector procList = new Vector(localProcs.size() + remoteProcs.size() + 1);
    // insert local PROCs
    Enumeration procEnum = localProcs.elements();
    while (procEnum.hasMoreElements())
      procList.addElement(new RemoteProc((Proc)procEnum.nextElement()));
    if(allProcs) {
      // insert a null separator between local and remote PROCs
      procList.addElement(null);
      // insert remote PROCs
      procEnum = remoteProcs.elements();
      while (procEnum.hasMoreElements())
        procList.addElement(new RemoteProc((RemoteProc)procEnum.nextElement()));
    }
    return procList;
  }

  /**
   * Returns the connection corresponding to the specified host,
   * or null if there is no such connection.
   * @param hostName  the host to return a connection for
   */
  public static Connection getConnection(String hostName) {
    int i;
    Connection connection = null;
    for(i=0; i<connections.size(); i++) {
      connection = (Connection)connections.elementAt(i);
      if(connection.hostname.equals(hostName)) {
        break;
      }
    }
    if(i==connections.size()) { // not found
      System.out.println("Engine is not connected to host " + hostName);
      return null;
    }
    return connection;
  }

  public static Object getProc(int procID) {
		Proc proc;
		if((proc = (Proc)localProcs.get(new Integer(procID))) != null)
			return proc;
		return (RemoteProc)remoteProcs.get(new Integer(procID));
  }

  public static Proc getLocalProc(int procID) {
		return (Proc)localProcs.get(new Integer(procID));
  }

  public static RemoteProc getRemoteProc(int procID) {
		return (RemoteProc)remoteProcs.get(new Integer(procID));
  }

  public static int getPort() {
		return port;
  }

  public static void displayGraph() {
    if (graph == null) {
      //System.out.println("graph is null");
      graph = new Graph("DACIA: application layout for host " + hostname);
      graph.show();
      graph.redrawAll();
    }
    else if (!graph.isShowing()) {
      graph.show();
      updateFlag = true;
      //updateGraph();
    }
  }
 
  static Object loadClass(String name, String fromhost) {	
    Class c;	
    Object obj = null;	
    int nconns, i=0;
    Connection connection;

    if (loader == null)
      loader = new DClassLoader();
    if (fromhost == null) 
      loader.setConnection(null);
    else {
      if ((nconns=connections.size()) > 0) {
        for (i=0; i<nconns; i++) {
          connection = (Connection)connections.elementAt(i);
          if (connection.hostname.equals(fromhost)) {
            loader.setConnection(connection);
            break;
          }
        }
      }
      if (i == nconns || nconns == 0) {
        System.out.println("loadClass(): No connection to" + fromhost);
        return null;
      }
    } 
    try {
      c = loader.loadClass (name);
    }
    catch (ClassNotFoundException e) {
      System.out.println("loadClass(): " + e.getClass().getName() + " " +
                         e.getMessage());
      e.printStackTrace();
      return null;
    }
    if (c != null) {
      try {
        obj = c.newInstance();
        System.out.println("loadClass(): create new instance of "+c.getName());
      }
      catch (Exception e) {
        System.out.println("loadClass(): newInstance() " +
                           e.getClass().getName() + " " + e.getMessage());
        e.printStackTrace();
      }
    }
    else 
      System.out.println("loadClass() unsuccessful - can't load class");
    return obj;
  }

}


/**
 * A thread used by the Engine to receive messages from another Engine or
 * to monitor the system and perform adaptation
 */
class EngineThread extends Thread {
  /* The network connection this thread listens to.
   * The connection is null for a ServerSocket
   */
  Connection connection = null;
  protected EngineThread(Connection conn) {
    connection = conn;
  }

  public void run() {
    if(connection == null) {
      Engine.listenToSocket();
    }
    else {
      Engine.listenToSocket(connection);
    }
  }
}


/**
 * A NetworkMessage represents the information exchanged accross the
 * network between two engines. The data field of the message can be
 * of any type, as specified by the type field.
 */
class NetworkMessage implements Serializable {
  /** The message type */
  int type = 0;
  /** The data in this message */
  int i1=0, i2=0, i3=0, i4=0, i5=0;
  Object data = null;

  public NetworkMessage(int msgType) {
    type = msgType;
  }

  /**
   * Constructs a message by reading data from the specified connection.
   * @param connection  the connection to read the message from
   */
  public NetworkMessage(Connection connection) throws Exception {
    DataInputStream input, dis;
    ObjectInputStream ois;
    byte buf[] = new byte[24];
    input = connection.getDIS();
    input.readFully(buf, 0, 24);
    dis = new DataInputStream(new ByteArrayInputStream(buf));
    type = dis.readInt();
    i1 = dis.readInt();
    i2 = dis.readInt();
    i3 = dis.readInt();
    i4 = dis.readInt();
    i5 = dis.readInt();
    if ((type == Engine.PROCLIST) || (type == Engine.MOVE) ||
        (type == Engine.DATA) || (type == Engine.LOAD)) {
      buf = new byte[i5];
      input.readFully(buf, 0, i5);
      ois = new ObjectInputStream(new ByteArrayInputStream(buf));
      //ObjectInputStream ois = connection.getOIS();	  	        
      try {
        data = ois.readObject();	  
      }
      catch (ClassNotFoundException e) {
        EngineThread tempThread = new EngineThread(connection);
        tempThread.start();
        Engine.loadClass(e.getMessage(), connection.hostname);
        tempThread.stop();
        ois = new ObjectInputStream(new ByteArrayInputStream(buf));
        data = ois.readObject();			 
      } 
    }
    else if (type == Engine.CLASS) {	  
     if (i5 > 0) {		
       data = new byte[i5];
       input.readFully((byte[])data, 0, i5);
     }
    }
    //System.out.println("RCVD");
  }

  /**
   * Sends this message on a specified connection.
   * @param connection  the connection where the message should be sent
   */
  public void sendMessage(Connection connection) {
    DataOutputStream output, dos;
    ByteArrayOutputStream baos, baos1 = null;
    ObjectOutputStream oos;
    byte buf[] = new byte[10000];
    try {
      output = connection.getDOS();
      synchronized(output) {
        baos = new ByteArrayOutputStream();
        dos = new DataOutputStream(baos);
        dos.writeInt(type);
        dos.writeInt(i1);
        dos.writeInt(i2);
        dos.writeInt(i3);
        dos.writeInt(i4);
        if((type == Engine.PROCLIST) || (type == Engine.MOVE) ||
           (type == Engine.DATA) || (type == Engine.LOAD)) {
          baos1 = new ByteArrayOutputStream();
          oos = new ObjectOutputStream(baos1);
          oos.writeObject(data);
          dos.writeInt(baos1.size());
          output.write(baos.toByteArray(), 0, 24);
          output.write(baos1.toByteArray(), 0, baos1.size());
          //output.write(buf, 0, 500);
          //ObjectOutputStream oos = connection.getOOS();
          //oos.writeObject(data);
          //oos.flush();
        }	
        else if (type == Engine.CLASS) {		  
          i5 = ((byte[])data).length;
          dos.writeInt(i5);
          output.write(baos.toByteArray(), 0, 24);
          output.write((byte[])data, 0, i5);		
        }
        else {
          dos.writeInt(i5);
          output.write(baos.toByteArray(), 0, 24);
        }
        output.flush();
      }
    }
    catch (Exception e) {
      System.out.println("Message: exception in sendMessage(): " +
                         e.getMessage());
      e.printStackTrace();
    }
    //System.out.println("SEND");
  }
}


/*
 * The local image of a remote Engine.
 * This class is not used in this version.
class EngineImage {
  // The hostname where this engine resides
  String hostname = null;
  // The connection to this engine, if any
  Connection connection = null;
  // The ID of the last message received from this engine
  int lastMsg = 0;
}
 */
