/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 2000 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: radu@eecs.umich.edu
*******************************************************************/

package dacia;

import java.io.Serializable;
import java.util.*;
import java.io.*;

/**
 * Implements the generic PROC, having a variable number of ports.
 * @see Engine
 * @see Port
 * @author Radu Litiu
 */
public abstract class Proc implements Serializable {
  /**
   * identifies the type of PROC: TRANSFORM, MERGE, etc.<P>
   * The types are defined in the subclasses corresponding to specific PROCs.
   */
  int type = NONE;
  public final static int NONE      = 0;
  public final static int TRANSFORM = 1;
  public final static int MERGE     = 2;
  public final static int REPLICATE = 3;
  public final static int FILTER    = 4;
  public final static int SELECT    = 5;
  /**
   * number of ports
   */
  public int portNo=0;
  /**
   * input/output ports
   * The ports are visible to the other PROCs.
   */
  public Port port[];
  /**
   * The number of asynchrounous messages queued, received on each port
	 */
  int msgNo[];
  /**
   * Identification fields
   */
  int id = 0;
  public String name = "Generic Proc";
  public String hostname = "noHost";
  /**
   * Worker thread used to handle asynchronous messages
	 */
  Worker worker = null;
  /**
   * message queue used to store messages received asynchronously
   */
  public Vector msgQueue = null;
  /**
   * Reference count for the number of threads processing a message
   * on this PROC, including the thread's asynchronous thread
   */
  int refcnt = 0;
  /**
   * flag set to true if this PROC is about  to move
   */
  public boolean moving = false;

  /** Constructs a new PROC
   * @param procName  the name of this PROC. Usually identifies the PROC type.
   * @param nports  the number of ports of this PROC
   */
  public Proc(String procName, int nports) {
    int i;
    name = procName;
    portNo = nports;
    port = new Port[nports];
    msgNo = new int[nports];
    for(i=0; i<nports; i++) {
      port[i] = new Port(this);
      msgNo[i] = 0;
		}
    msgQueue = new Vector();
    id = Engine.getNextID();
    hostname = Engine.getHostName();
    worker = new Worker(this);
    worker.start();
  }

  /**
   * Generic method, has to be implemented in the subclass. <P>
   * Used for synchronous messages.
   */
  public abstract void handleMessage(Message msg, int portNo);

  /**
   * Wrapper for the handleAsyncMessage() method.
   * Increments/decrements the reference counter at the beginning/end
   */
  void handleAsyncMsg() {
    refcnt++;
    handleAsyncMessage();
    refcnt--;
  }

  /**
   * Generic method, has to be implemented in the subclass.
   * Used for asynchronous messages.
   * Should use the message returned by getMessage().
   * The thread should wait when the message queue is empty.
   * If the PROC is moving, this method should terminate and return.
   * Checking the length of the message queue(s) should be synchronized
   * on the message queue.
   */
  public abstract void handleAsyncMessage();

  /**
   * Inserts an asynchronous message in the message queue and notifies
   * the worker thread to handle the message.
   */
  void asyncMessageReceived(Message msg, int portNo) {
    msg.setPort(portNo);
    synchronized(msgQueue) {
      msgQueue.addElement(msg);
      msgNo[portNo]++;
      try {
        msgQueue.notify();
        //System.out.println(Thread.currentThread().toString()+" called notify");
      }
      catch (IllegalMonitorStateException e) {
        System.out.println(name + ": exception in notify()" + e.getMessage());
        e.printStackTrace();
      }
    }
  }

  /**
   * Sends a message through the specified port of this PROC.<p>
   * synchronous = 1 for synchronous communication, 0 otherwise.
	 * Asynchronous output is recommended when the receiving PROC has
	 * multiple interfaces and it has to synchronize the data received
	 * on these interfaces.
   */
  public void output(int portNo, Message msg, int synchronous) {
    Proc proc = null;
    Port thePort = port[portNo];
    if(thePort.connectionStatus == 1) {
      // XXX - !!! proc can be made null in the meantime
      if((proc = thePort.connectedProc) == null) {
        if(thePort.remoteConnection != null) {
          Engine.sendMessage(msg, thePort.remoteConnection,
                             thePort.connectedProcID, thePort.connectedPortNo);
        }
        else {
          System.out.println("Can't send a message: connected proc null: "
                           + id + ":" + portNo);
        }
      }
      else if(proc.moving) {
        synchronized(proc) {
          try {
            proc.wait();
          }
          catch (Exception e) {
            System.out.println("output(): exception raised while wait(): "
                               + e.getMessage());
            e.printStackTrace();
          }
        }
        if(thePort.connectionStatus == 2) {
          Engine.sendMessage(msg, thePort.remoteConnection,
                             thePort.connectedProcID, thePort.connectedPortNo);
        }
        else {
          System.out.println(
            "!!! connectionStatus should have been 2 after this PROC movement "
            + id + ":" + portNo);
        }
      }
      else {
        proc.input(thePort.connectedPortNo, msg, synchronous);
      }
    }
    else if(thePort.connectionStatus == 2) {
      Engine.sendMessage(msg, thePort.remoteConnection,
                         thePort.connectedProcID, thePort.connectedPortNo);
    }
    else {
      System.out.println("Can't send message, port is not connected "
                         + id + ":" + portNo);
    }
  }

  /**
   * Sends a message through the specified port of this PROC.<p>
   * The output type (synchronous or asynchronous is given by the default
   * type of the output port.
   */
  public void output(int portNo, Message msg) {
    output(portNo, msg, port[portNo].getOutputType());
  }

  /**
   * Receives a message through the specified port.<P>
   * synchronous = 1 for synchronous communication, 0 otherwise
   */
  void input(int portNo, Message msg, int synchronous) {
    refcnt++;
    if(synchronous == 1) {
      handleMessage(msg, portNo);
    }
    else {
      asyncMessageReceived(msg, portNo);
    }
    refcnt--;
  }

  /**
   * Returns the first message in the queue and removes it from the queue.<p>
   * Decrements the counter for a port.
   * If the PROC is about to move, returns null.
	 */
  public Message getMessage() {
    Message msg = null;
    synchronized(msgQueue) {
      while(msgQueue.isEmpty() && !moving) {
        //System.out.println(Thread.currentThread().toString()+" waiting");
        try {
          msgQueue.wait();
        }
        catch (Exception e) {
          System.out.println(name + ": exception raised in wait(): "
                             + e.getMessage());
          e.printStackTrace();
        }
      }
      if(moving)  return null;
      // remove() is java2 specific. java2 currently has problems using
      // awt on my debian box. Will get back to this when this problem
      // is solved.
      //msg = (Message)msgQueue.remove(0);
      msg = (Message)msgQueue.elementAt(0);
      msgQueue.removeElementAt(0);
      msgNo[msg.getPort()]--;
    }
    return msg;
  }

  /**
   * Returns the first message in the queue.<p>
   * Decrements the counter for a port if needed.
   * This method is not currently used.
	 */
  public Message getMessage(boolean remove) {
    synchronized(msgQueue) {
      Message msg = (Message)msgQueue.elementAt(0);
      if(remove) {
        msgQueue.removeElementAt(0);
        msgNo[msg.getPort()]--;
      }
      return msg;
    }
  }

  /**
   * Returns the first message in the queue that has been received on
   * the port indicated. Decrements the counter for the port if needed.
	 */
  public Message getMessage(int port, boolean remove) {
    Message msg = null;
    int i, size;
    synchronized(msgQueue) {
      size = msgQueue.size();
      for(i=0; i<size; i++) {
        msg = (Message)msgQueue.elementAt(i);
        if(msg.getPort() == port) {
          if(remove) {
            msgQueue.removeElementAt(i);
            msgNo[port]--;
          }
          return msg;
        }
      }
      return null;
    }
  }

  public int getID() {
    return id;
  }

  /**
   * Prints information about this Proc and its connections.
   */
  public void printInfo() {
    System.out.println("    " + id + " " + name + ":" + type + ", " +
                       portNo + " ports, host: " + hostname);
    for(int i=0; i<portNo; i++) {
      if(port[i].connectionStatus != 0) {
        System.out.println("\t" + i + " -> " + port[i].connectedProcID +
                           ":" + port[i].connectedPortNo);
      }
    }
  }

  /**
   * Prints information about this Proc and its connections to the
   * indicated output stream.
   * @param dos  the output stream to print the info to
   */
  public void printInfo(DataOutputStream dos) {
    try{
      dos.writeBytes("    " + id + " " + name + ":" + type + ", " +
                     portNo + " ports, host: " + hostname + "\n");
      for(int i=0; i<portNo; i++)
        if(port[i].connectionStatus != 0)
          dos.writeBytes("\t" + i + " -> " + port[i].connectedProcID + ":" +
                         port[i].connectedPortNo + "\n");
    }
    catch(IOException e) {
      System.out.println("Proc: exception in printInfo: " + e.getMessage());
      e.printStackTrace();
    }
  }

  /**
   * Starts an action on this PROC.<p>
   * By default it does nothing. Customized actions should be declared
   * in the subclass.
   */
  public void start() { }

  /**
   * Prepares this PROC for moving to another Engine.<p>
   * Attempts to free as many objects as possible from this PROC, so that
   * they are not serialized. Only the data representing state particular
   * to this PROC should be kept. There is no need to submit general objects
   * that can be restored at the destination by calling their constructor.
   * This method needs to be overwritten in the subclass, to handle the
   * customized structure of the subclass object. The better the customized
   * method is written, the more efficient the serialization is.
   * For PROCs with GUI, this method should dispose the GUI objects.
   */
  public void pack() { }

  /**
   * Restores the state of this PROC after moving from another Engine.<p>
   * This method needs to be overwritten in the subclass.
   * This method is the inverse of the pack() method.
   */
  public void unpack() { }

  /**
   * Start the worker thread
   */
  void startWorker() {
    worker = new Worker(this);
    worker.start();
  }

  /**
   * Stop the worker thread
   */
  void stopWorker() {
    /**
     * Killing the worker thread may cause problems due to messages in traffic.
     * I guess it is solved now, but ...
     */
    try {
      worker.stop();
    }
    catch (SecurityException e) {
      System.out.println("Can't stop thread: " + e.getMessage());
      e.printStackTrace();
    }
    worker = null;
  }

  /**
   * return the first available port
   * @return the available port if successful, -1 otherwise
   */
  public int getFreePort() {
    for(int i=0; i<portNo; i++) {
      if(port[i].connectionStatus == 0) {
        return i;
      }
    }
    System.out.println("There are no free ports on this Resource");
    return -1;
  }

  public String getName() {
    return name;
  }
}
          
/**
 * The thread that handles the messages received asynchronously
 */
class Worker extends Thread {
  private Proc myProc  = null;

  protected Worker(Proc proc) {
    myProc = proc;
  }

  public void run() {
    myProc.handleAsyncMsg();
 }
}

