/**
 * This class is modified from "Java Examples in a Nutshell". 
 */

package dacia.apps.wb;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;

/** 
 * This class creates and displays a window containing a list of
 * files and sub-directories in a specified directory.  Clicking on an
 * entry in the list displays more information about it. Double-clicking 
 * on an entry displays it, if a file, or lists it if a directory.
 * An optionally-specified FilenameFilter filters the displayed list.
 **/
public class FileLister extends Frame 
                        implements ActionListener, ItemListener {
  private List list;                   // To display the directory contents in
  private Label location;          // To display detail info in.
  private Panel buttons;               // Holds the buttson
  private Choice filterlist;
  private Button up, close;            // The Up and Close buttons
  private File currentDir;             // The directory currently listed
  private FilenameFilter filter;       // An optional filter for the directory
  private String[] files;              // The directory contents
  
  Whiteboard target;
  
  /** 
   * Constructor: create the GUI, and list the initial directory. 
   **/
  public FileLister(Whiteboard target, String directory) { 
    super("File Lister");              // Create the window
    this.filter = filter;              // Save the filter, if any
    this.target = target;

    list = new List(12, false);        // Set up the list
    list.setFont(new Font("MonoSpaced", Font.PLAIN, 14));
    list.addActionListener(this);
    list.addItemListener(this);
    
    location = new Label();         // Set up the location area
    location.setFont(new Font("MonoSpaced", Font.PLAIN, 12));

    buttons = new Panel();             // Set up the button box
    buttons.setLayout(new FlowLayout(FlowLayout.RIGHT, 15, 5));
    buttons.setFont(new Font("SansSerif", Font.BOLD, 14));

	filterlist = new Choice();
	filterlist.addItemListener(this);
	filterlist.addItem("All image files (*.*)");
	filterlist.addItem("*.gif");
	filterlist.addItem("*.jpg");
	filterlist.addItem("*.png");
    buttons.add(filterlist);             

    up = new Button("Up a Directory"); // Set up the two buttons
    up.addActionListener(this);
    buttons.add(up);                   // Add buttons to button box

    close = new Button("Close");
    close.addActionListener(this);
    buttons.add(close);
    
    this.add(list, "Center");          // Add stuff to the window
    this.add(location, "North");
    this.add(buttons, "South");
    this.setSize(500, 350);
    
    listDirectory(directory);          // And now list the initial directory.
	show();
  }
  
  /**
   * This method uses the list() method to get all entries in a directory
   * and then displays them in the List component. 
   **/
  public void listDirectory(String directory) {
    // Convert the string to a File object, and check that the dir exists
    File dir = new File(directory);
    if (!dir.isDirectory()) 
      throw new IllegalArgumentException("FileLister: no such directory");

    // Get the (filtered) directory entries
    files = dir.list(filter);        
    // And sort them.
    Sorter.sort(files);

    // Remove any old entries in the list, and add the new ones
    list.removeAll();
    list.addItem("..");  // A special case entry
    for(int i = 0; i < files.length; i++) list.addItem(files[i]);
    
    // Display directory name in window titlebar and in the location box
    this.setTitle(directory);
    location.setText(directory);

    // Remember this directory for later.
    currentDir = dir;
  }
  
  /**
   * This ItemListener method uses various File methods to obtain information
   * about a file or directory. Then it displays that info in the location box.
   **/
  public void itemStateChanged(ItemEvent e) {
	  if (e.getSource() == list) {
		  int i = list.getSelectedIndex() - 1;      // minus 1 for Up To Parent entry
		  if (i < 0) return;
		  String filename = files[i];               // Get the selected entry 
		  File f = new File(currentDir, filename);  // Convert to a File
		  if (!f.exists())                          // Confirm that it exists
			  throw new IllegalArgumentException("FileLister: " +
                                         "no such file or directory");

		  // Get the location about the file or directory, concatenate to a string
		  String info = f.getAbsolutePath();
		  if (f.isDirectory()) info += File.separator;
		  // And display the location string
		  location.setText(info);
	  }
	  else if (e.getSource() == filterlist) {
		  int i = filterlist.getSelectedIndex();
		  if (i == 0)
			  filter = null;
		  else {
			  final String suffix = filterlist.getSelectedItem().substring(2);
			  filter = new FilenameFilter() {
			    public boolean accept(File dir, String name) {
				  if (name.endsWith(suffix)) return true;
				  else return (new File(dir, name)).isDirectory();				
				}
			  };
		  }
		  listDirectory(currentDir.getAbsolutePath());
	  }  
  }

  /**
   * This ActionListener method is invoked when the user double-clicks on an 
   * entry or clicks on one of the buttons.  If they double-click on a file,
   * create a FileViewer to display that file.  If they double-click on a
   * directory, call the listDirectory() method to display that directory
   **/
  public void actionPerformed(ActionEvent e) {
    if (e.getSource() == close) this.dispose();
    else if (e.getSource() == up) { up(); }
    else if (e.getSource() == list) {  // Double click on an item
      int i = list.getSelectedIndex(); // Check which item
      if (i == 0) up();                // Handle first Up To Parent item
      else {                           // Otherwise, get filename
        String name = files[i-1]; 
        File f = new File(currentDir, name);    // Convert to a File
        String fullname = f.getAbsolutePath();
        if (f.isDirectory()) listDirectory(fullname);  // List a directory
		else {
			URL url = null;
			try {
				url = f.toURL();
			}
			catch(MalformedURLException excp) {
				dispose();
			}
			target.frame.panel.setImage(url);
			dispose();
		}
      }
    }
  }

  /** A convenience method to display the contents of the parent directory */
  protected void up() {
    String parent = currentDir.getParent();
    if (parent == null) return;
    listDirectory(parent);
  }

}

